from datetime import datetime
import os

import numpy as np
import torch
from torch import nn, flatten

from data.cabackend import Cabackend
from data.wesad import Wesad
from helpers.soc_comm import ClientBuffer

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"
from torch.utils.data import DataLoader

WANDB_GROUP = "noise_0004"
# DROPOUT = 0.1
LEARNING_RATE = 0.0001
BATCH_SIZE = 2048
EPOCHS = 250
EPCH_MULT = 3
DROPOUT = 0.3

NEGATIVE_WEIGHT = -1
POSITIVE_WEIGHT = 0.1

DATA_SEED = 55796
METHOD = "ours_to"

ENC_LSTM = 609
ENC_DENSE = 609

CLASS_01 = 64
CLASS_02 = 32

NOISE_TYPE = "laplace"
# Increase this by 4x every next run
NOISE_SCALE = 32
DATASET = "cabackend"
# DATASET = "wesad"


class Encoder(nn.Module):
    def __init__(self, input_size=6):
        super(Encoder, self).__init__()
        self.lstm = nn.LSTM(input_size=input_size, hidden_size=ENC_LSTM)
        self.dropout = nn.Dropout(DROPOUT)
        self.linear_01 = nn.Linear(ENC_DENSE * 7, ENC_DENSE)
        # self.batch_norm = nn.BatchNorm1d(h1)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.lstm(x)
        x = flatten(x[0], start_dim=1)
        x = self.dropout(x)
        x = self.linear_01(x)
        # x = self.batch_norm(x)
        x = self.relu(x)
        return x


class BaselineNoise:
    def __init__(self, trainset, validset, testset, noise_type):
        self.trainset = trainset
        self.validset = validset
        self.testset = testset
        self.run_times = []
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        assert noise_type in ("laplace", "gauss")
        self.noise_type = noise_type

        self.i = 0
        # Define model
        # self.model_encoder = Encoder().float().to(self.device)
        # Define optimizer
        # self.optimizer = torch.optim.Adam(
        #     self.model_encoder.parameters(), lr=LEARNING_RATE
        # )
        self.stop_check = None
        self.buffer = ClientBuffer()

    def _add_noise(self, X):
        if self.noise_type == "laplace":
            noised = X + np.random.laplace(size=X.shape, scale=NOISE_SCALE)
        elif self.noise_type == "gauss":
            noised = X + np.random.normal(size=X.shape, scale=NOISE_SCALE)
        else:
            raise ValueError("Invalid noise type selected.")
        # Reshape to fit in the server classifiers.
        noised = torch.flatten(noised, start_dim=1)
        return noised.to(torch.float32)

    def train(self):
        print("Using {} device.".format(self.device))
        # self.model_encoder.summary()
        adv_epocs = EPOCHS * EPCH_MULT

        # Original PAN
        if METHOD == "pan":
            # Not utilised here, but an easy fix for missing mister stop in reporting.
            self.pan_step()
        # Our PAN
        elif "ours" in METHOD:
            for i in range(8):
                print("Loop #{}".format(i + 1))
                if not self.normal_step():
                    break
                self.adversary_step(epochs=adv_epocs)
        else:
            raise ValueError("Invalid method selected.")
        self.buffer.send_forward("MODE_OUT")
        self.adversary_step(epochs=5000)
        self.buffer.send_forward("MODE_STOP")
        print("Training complete.")

    def normal_step(self):
        print("NORMAL MODE")
        # Train the network as specified, batch by batch
        for e in range(EPOCHS):
            self.i += 1
            s = datetime.now()
            # self.model_encoder.train()

            # Steps 1 & 2
            self.buffer.send_forward("MODE_STEP1")
            for X, y in self.trainset:
                X = self._add_noise(X).float()
                y_user = y[0].float()
                y_task = y[1].float()
                if not self.step_01(X, y_user):
                    return False
                self.step_02(X, y_task)

            # Step 3
            self.buffer.send_forward("MODE_STEP3")
            for X, y in self.trainset:
                X = self._add_noise(X).float()
                y_user = y[0].float()
                y_task = y[1].float()
                self.step_03(X, y_user, y_task)

            # Testing
            # self.model_encoder.eval()
            self.buffer.send_forward("MODE_TEST_TRAIN")
            # print("Testing on trainset.")
            for X, y in self.trainset:
                X = self._add_noise(X).float()
                y_user = y[0].float()
                y_task = y[1].float()
                self.step_test(X, y_user, y_task)
            self.buffer.send_forward("MODE_TEST_VALID")
            # print("Testing on validset.")
            for X, y in self.validset:
                X = self._add_noise(X).float()
                y_user = y[0].float()
                y_task = y[1].float()
                self.step_test(X, y_user, y_task)
            # print("Testing complete.")

            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))
        return True

    def adversary_step(self, epochs=EPOCHS):
        print("ADVERSARY MODE")
        # self.model_encoder.eval()
        # Only do this when retraining the adversary classifier
        for e in range(epochs):
            self.buffer.send_forward("MODE_STEPADV")
            torch.cuda.empty_cache()
            self.i += 1
            s = datetime.now()

            for X, y in self.trainset:
                X = self._add_noise(X).float()
                y_task = y[1].float()
                self.step_02(X, y_task)

            # Testing
            self.buffer.send_forward("MODE_TEST_TRAIN")
            for X, y in self.trainset:
                X = self._add_noise(X).float()
                y_user = y[0].float()
                y_task = y[1].float()
                self.step_test(X, y_user, y_task)
            self.buffer.send_forward("MODE_TEST_VALID")
            # print("Testing on validset.")
            for X, y in self.validset:
                X = self._add_noise(X).float()
                y_user = y[0].float().to(self.device)
                y_task = y[1].float().to(self.device)
                self.step_test(X, y_user, y_task)

            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))

        # TODO: finish the implementation of this (also server-side) to get testing results.
        for X, y in self.validset:
            X = self._add_noise(X).float()
            y_user = y[0].float().to(self.device)
            y_task = y[1].float().to(self.device)
            self.step_test(X, y_user, y_task)

    def step_01(self, X, y_user):
        """
        Training of the good classifier.
        """
        # logits = self.model_encoder(X)
        logits = torch.flatten(X, start_dim=1)
        # print("Sending user.")
        self.buffer.send_forward(logits)
        self.buffer.send_forward(y_user)

        X_grad = self.buffer.take_backward()
        if X_grad == "MODE_ENDADV":
            print("Training not progressing, switching to the last adversarial mode.")
            return False
        return True

    def step_02(self, X, y_task):
        """
        Training of the bad classifier.
        The encoder does not get updated with feedback from the bad classifier,
        so this step is as simple as calculating logits and passing them to
        the server, alongside with the task labels.
        """
        # logits = self.model_encoder(X)
        logits = torch.flatten(X, start_dim=1)
        self.buffer.send_forward(logits)
        self.buffer.send_forward(y_task)
        # print("Sending task.")

    def step_03(self, X, y_user, y_task):
        """
        Training of the whole system with weighted loss functions (on the server side).
        """
        # logits = self.model_encoder(X)
        logits = torch.flatten(X, start_dim=1)
        # print("Sending both.")
        self.buffer.send_forward(logits)
        self.buffer.send_forward(y_user)
        self.buffer.send_forward(y_task)

        # The first grad we receive here is still from step 1, as server sends out the grad every
        # epoch, while on the client side we skip the first epoch in terms of backpropagation.
        X_grad = self.buffer.take_backward()

    def step_test(self, X, y_user, y_task):
        """
        Send data through the encoder for the server to do the tests.
        """
        # logits = self.model_encoder(X)
        logits = torch.flatten(X, start_dim=1)

        self.buffer.send_forward(logits)
        self.buffer.send_forward(y_user)
        self.buffer.send_forward(y_task)
        # print("Sending test.")


if __name__ == "__main__":
    if DATASET == "wesad":
        wesad = Wesad(
            data_dir="/wesad/",
            seed=DATA_SEED,
            test=False,
        )
        train_data, validation_data, test_data = wesad.get_lstm_data()
    elif DATASET == "cabackend":
        cabackend = Cabackend()
        train_data, validation_data, test_data = cabackend.get_lstm_data()
    else:
        raise ValueError("Invalid dataset set.")

    train_loader = DataLoader(
        train_data, batch_size=BATCH_SIZE, shuffle=True, drop_last=True
    )
    valid_loader = DataLoader(
        validation_data, batch_size=BATCH_SIZE, shuffle=True, drop_last=True
    )
    test_loader = DataLoader(
        test_data, batch_size=BATCH_SIZE, shuffle=False, drop_last=False
    )

    model = BaselineNoise(train_loader, valid_loader, test_loader, NOISE_TYPE)
    model.train()
