import os
from random import shuffle, sample

import numpy as np
import torch
from numpy import repeat, array
from pandas import read_csv, concat, DataFrame, unique
from sklearn.preprocessing import StandardScaler
from torch import tensor, nn
from torch.nn.functional import one_hot
from torch.utils.data import Dataset, random_split

import settings


class Cabackend:
    def __init__(
        self,
        split=[0.7, 0.15, 0.15],
        meta=["user", "seance"],
        # features=["ax_me", "ax_sd", "ax_mcr", "ax_mai", "ay_me", "ay_sd", "ay_mcr", "ay_mai", "az_me", "az_sd",
        #           "az_mcr", "az_mai", "a_me", "a_sd", "a_mcr", "a_mai", "gx_me", "gx_sd", "gx_mcr", "gy_me",
        #           "gy_sd", "gy_mcr", "gz_me", "gz_sd", "gz_mcr", "g_me", "g_sd", "g_mcr", "fa_me", "fa_sd",
        #           "fa_mcr", "fb_me", "fb_sd", "fb_mcr", "fc_me", "fc_sd", "fc_mcr", "fd_me", "fd_sd",
        #           "fd_mcr"],
        features=[
            "ax_me",
            "ax_sd",
            "ax_mcr",
            "ax_mai",
            "ay_me",
            "ay_sd",
            "ay_mcr",
            "ay_mai",
            "az_me",
            "az_sd",
            "az_mcr",
            "az_mai",
            "a_me",
            "a_sd",
            "a_mcr",
            "a_mai",
            "gx_me",
            "gx_sd",
            "gx_mcr",
            "gy_me",
            "gy_sd",
            "gy_mcr",
            "gz_me",
            "gz_sd",
            "gz_mcr",
            "g_me",
            "g_sd",
            "g_mcr",
            "fa_me",
            "fa_sd",
            "fa_mcr",
            "fb_me",
            "fb_sd",
            "fb_mcr",
            "fc_me",
            "fc_sd",
            "fc_mcr",
            "fd_me",
            "fd_sd",
            "fd_mcr",
            "ca_me",
            "ca_sd",
            "ca_min",
            "ca_max",
            "ca_mcr",
            "cb_me",
            "cb_sd",
            "cb_min",
            "cb_max",
            "cb_mcr",
            "cc_me",
            "cc_sd",
            "cc_min",
            "cc_max",
            "cc_mcr",
            "cd_me",
            "cd_sd",
            "cd_min",
            "cd_max",
            "cd_mcr",
            "c_me",
            "c_sd",
            "c_min",
            "c_max",
            "c_mcr",
            "m_me",
            "m_sd",
            "m_min",
            "m_max",
            "m_jc",
            "m_jr",
            "m_jv",
            "m_iji",
            "ns_me",
            "ns_sd",
            "ns_sum",
            "ns_jc",
            "ns_jr",
            "ns_jv",
            "ns_iji",
            "nr_me",
            "nr_sd",
            "nr_sum",
            "nr_jc",
            "nr_jr",
            "nr_jv",
            "nr_iji",
        ],
        data_path="data/cabackend/",
        userlim=None,
    ):
        self.split = split
        self.meta = meta
        self.features = features
        self.data_path = data_path

        self.train_seances = []
        self.test_seances = []

        self.users = settings.CABACKEND_USERS
        if userlim:
            self.users = sample(self.users, userlim)
            print("Taking a user subset: {}".format(self.users))

        self.train = None
        self.validation = None
        self.test = None
        self.X_train = None
        self.y_train = None
        self.X_valid = None
        self.y_valid = None
        self.X_test = None
        self.y_test = None

    def get_lstm_data(self):
        """
        Generate data of shape (N, L, H_in) where;
        N -> batch size
        L -> sequence length (lookback)
        H_in -> input size (number of features
        """
        lookback = settings.LOOKBACK
        data = concat(
            [
                read_csv(self.data_path + "task_1_interval_1.csv"),
                read_csv(self.data_path + "task_2_interval_1.csv"),
                read_csv(self.data_path + "task_3_interval_1.csv"),
            ]
        )

        # Subsample users and features
        data = data[data["user"].isin(self.users)]
        data = data[self.meta + self.features]
        # Take only a single session from each user
        seances = sorted(unique(data["seance"]))
        test_data = data[
            data["seance"].isin(
                [seances[i] for i in [x for x in range(len(seances)) if x % 6 < 3]]
            )
        ]
        train_data = data[
            data["seance"].isin(
                [seances[i] for i in [x for x in range(len(seances)) if x % 6 >= 3]]
            )
        ]
        X, y_utility, y_privacy = self.lstm_norm_and_split(train_data, lookback)
        train_dataset = CabackendDataset(X, y_utility, y_privacy)

        X, y_utility, y_privacy = self.lstm_norm_and_split(test_data, lookback)
        valid_dataset, test_dataset = random_split(
            CabackendDataset(X, y_utility, y_privacy),
            [0.5, 0.5],
            torch.Generator().manual_seed(42),
        )

        # X, y_utility, y_privacy = self.lstm_norm_and_split(data, lookback)
        # train_dataset, valid_dataset, test_dataset = random_split(
        #     CabackendDataset(X, y_utility, y_privacy),
        #     self.split,
        #     torch.Generator().manual_seed(42),
        # )

        return train_dataset, valid_dataset, test_dataset

    def lstm_norm_and_split(self, data, lookback):
        data = data.fillna(0)

        # Data normalisation
        for feature in self.features:
            data[feature] = (data[feature] - data[feature].mean()) / data[feature].std()
        data = data.fillna(0)

        # Construct the required data
        X, y_utility, y_privacy = self.lstm_construct_single(data, lookback)

        X = tensor(array(X)).float()
        y_utility = one_hot(tensor(array(y_utility))).float()
        y_privacy = one_hot(tensor(array(y_privacy))).float()

        return X, y_utility, y_privacy

    def lstm_construct_single(self, data, lookback):
        X = []
        y_utility = []
        y_privacy = []

        # Construct the lookback tensor
        seance_i = 0
        for seance in sorted(unique(data["seance"])):
            seance_data = data[data["seance"] == seance]
            user = self.users.index(seance_data.iloc[0, 0])
            seance = seance_i % 3

            for i in range(lookback, len(seance_data)):
                x = list(seance_data[self.features].iloc[i - lookback : i, :].values)
                X.append(x)
                y_utility.append(user)
                y_privacy.append(seance)
            seance_i += 1

        return X, y_utility, y_privacy


class CabackendDataset(Dataset):
    def __init__(self, data, users, activities):
        self.data = data
        self.users = users
        self.activities = activities

    def __len__(self):
        return len(self.users)

    def __getitem__(self, item):
        return self.data[item], [self.users[item], self.activities[item]]


if __name__ == "__main__":
    rnn = nn.LSTM(10, 20, 2, batch_first=True)
    input = torch.randn(3, 5, 10)
    h0 = torch.randn(2, 3, 20)
    c0 = torch.randn(2, 3, 20)
    output, (hn, cn) = rnn(input, (h0, c0))

    print(output.shape)
