import os
from collections import Counter
from warnings import filterwarnings

import numpy as np
import pandas
from pandas import DataFrame
from pandas.core.common import SettingWithCopyWarning
from torch import tensor
from torch.nn.functional import one_hot

import settings
from data.cabackend import CabackendDataset

filterwarnings(action="ignore", category=SettingWithCopyWarning)


class Opportunity:
    def __init__(
        self,
        data_dir="opportunity/OpportunityUCIDataset/raw/",
        column_dir="opportunity/OpportunityUCIDataset/dataset/column_names.txt",
        activity_col="Locomotion",
    ):
        self.data_dir = data_dir
        self.column_dir = column_dir

        self.column_names = self.get_column_names()
        self.column_data = self.column_names[:242]
        self.column_data = [
            "Accelerometer RKN^ accX",
            "Accelerometer RKN^ accY",
            "Accelerometer RKN^ accZ",
            "Accelerometer HIP accX",
            "Accelerometer HIP accY",
            "Accelerometer HIP accZ",
            "Accelerometer LUA^ accX",
            "Accelerometer LUA^ accY",
            "Accelerometer LUA^ accZ",
            "Accelerometer RUA_ accX",
            "Accelerometer RUA_ accY",
            "Accelerometer RUA_ accZ",
            "Accelerometer LH accX",
            "Accelerometer LH accY",
            "Accelerometer LH accZ",
            "Accelerometer BACK accX",
            "Accelerometer BACK accY",
            "Accelerometer BACK accZ",
            "Accelerometer RKN_ accX",
            "Accelerometer RKN_ accY",
            "Accelerometer RKN_ accZ",
            "Accelerometer RWR accX",
            "Accelerometer RWR accY",
            "Accelerometer RWR accZ",
            "Accelerometer RUA^ accX",
            "Accelerometer RUA^ accY",
            "Accelerometer RUA^ accZ",
            "Accelerometer LUA_ accX",
            "Accelerometer LUA_ accY",
            "Accelerometer LUA_ accZ",
            "Accelerometer LWR accX",
            "Accelerometer LWR accY",
            "Accelerometer LWR accZ",
            "Accelerometer RH accX",
            "Accelerometer RH accY",
            "Accelerometer RH accZ",
        ]
        self.column_meta = self.column_names[242:]
        self.activity_col = activity_col
        self.seance_split = ([0, 1, 2], [3], [4])

        self.users = ["1", "2", "3", "4"]
        self.lookback = settings.LOOKBACK

    def get_column_names(self):
        result = []
        with open(self.column_dir) as column_file:
            row = column_file.readline()
            while row:
                if row.startswith("Column"):
                    result.append(
                        " ".join(
                            row.split(":")[-1].split(";")[0].split(" ")[2:]
                        ).strip()
                    )
                row = column_file.readline()
        return result

    def import_data(self):
        data = DataFrame()
        users, seance_i, seances, = (
            [],
            1,
            [],
        )
        for data_file in sorted(os.listdir(self.data_dir)):
            new_data = pandas.read_table(
                self.data_dir + data_file, header=None, sep=" "
            )
            new_data.columns = self.column_names
            new_data = self.subsample_to_second(new_data)
            data = pandas.concat((data, new_data))
            user = data_file[1]
            users.extend(np.repeat(user, len(new_data)))
            seances.extend(np.repeat(seance_i, len(new_data)))
            seance_i += 1

        data.columns = self.column_names
        data = data.reset_index(drop=True)
        data["user"] = users
        data["seance"] = seances
        self.column_names = self.column_names + ["user", "seance"]
        self.column_meta = self.column_meta + ["user", "seance"]

        return data

    def subsample_to_second(self, data):
        # The data is sampled to 30 Hz, therefore we take every 30th record from the data.
        data = data.iloc[::15]
        return data

    def get_lstm_data(self):
        """
        Generate data of shape (N, L, H_in) where;
        N -> batch size
        L -> sequence length (lookback)
        H_in -> input size (number of features
        """
        # Import data from files
        data = self.import_data()

        # Subsample users and features
        data = data[data["user"].isin(self.users)]
        data = data[self.column_data + ["user", "seance", self.activity_col]]
        data = data.rename(columns={self.activity_col: "task"})
        data = data.fillna(0)

        # Normalise the data
        data = self.normalise(data)

        # Construct the data to LSTM input form
        (
            X_train,
            yu_train,
            yp_train,
            X_valid,
            yu_valid,
            yp_valid,
            X_test,
            yu_test,
            yp_test,
        ) = self.lstm_construct_single(data)

        print("User counters:")
        print(Counter(yu_train))
        print(Counter(yu_valid))
        print(Counter(yu_test))
        print()
        print("Task counters:")
        print(Counter(yp_train))
        print(Counter(yp_valid))
        print(Counter(yp_test))

        # Convert to tensors and one-hot encode target variables
        X_train, yu_train, yp_train = self.tensorise(X_train, yu_train, yp_train)
        X_valid, yu_valid, yp_valid = self.tensorise(X_valid, yu_valid, yp_valid)
        X_test, yu_test, yp_test = self.tensorise(X_test, yu_test, yp_test)
        print()
        print("Data shape:")
        print(X_train.shape, yu_train.shape, yp_train.shape)
        print(X_valid.shape, yu_valid.shape, yp_valid.shape)
        print(X_test.shape, yu_test.shape, yp_test.shape)

        # Return torch datasets
        return (
            CabackendDataset(X_train, yu_train, yp_train),
            CabackendDataset(X_valid, yu_valid, yp_valid),
            CabackendDataset(X_test, yu_test, yp_test),
        )

    def normalise(self, data):
        data["user_seance"] = (data["seance"] - 1) % 6
        train_data = data[data["user_seance"] < self.seance_split[1][0]]
        valid_data = data[data["user_seance"] == self.seance_split[1][0]]
        test_data = data[data["user_seance"] == self.seance_split[2][0]]

        # Data normalisation
        for feature in self.column_data:
            train_data[feature] = (
                train_data[feature] - train_data[feature].mean()
            ) / train_data[feature].std()
            valid_data[feature] = (
                valid_data[feature] - valid_data[feature].mean()
            ) / valid_data[feature].std()
            test_data[feature] = (
                test_data[feature] - test_data[feature].mean()
            ) / test_data[feature].std()
        data = pandas.concat((train_data, valid_data, test_data))
        data = data.fillna(0)
        return data

    def lstm_construct_single(self, data):
        X_train, X_valid, X_test = [], [], []
        yu_train, yu_valid, yu_test = [], [], []
        yp_train, yp_valid, yp_test = [], [], []

        for seance in pandas.unique(data["seance"]):
            if (seance - 1) % 6 == 5:
                continue
            seance_data = data[data["seance"] == seance]
            user = pandas.unique(seance_data["user"])
            if len(user) > 1:
                raise ValueError("More than one user in a single session.")
            # We start with 0 here, so one hot works as intended later on
            user = int(user[0]) - 1
            # Construct the lookback tensor
            for i in range(self.lookback, len(seance_data)):
                # We check if all privacy labels belong to the same class
                if (
                    len(pandas.unique(seance_data["task"].iloc[i - self.lookback : i]))
                    > 1
                ):
                    continue
                x = list(
                    seance_data[self.column_data].iloc[i - self.lookback : i, :].values
                )
                task = seance_data["task"].iloc[i]
                # Index 0 annotates none of the actions
                if task == 0:
                    continue
                task -= 1
                # Task with index 3 is missing
                if task > 2:
                    task -= 1
                if (seance - 1) % 6 < self.seance_split[1][0]:
                    X_train.append(x)
                    yu_train.append(user)
                    yp_train.append(task)
                elif (seance - 1) % 6 == self.seance_split[1][0]:
                    X_valid.append(x)
                    yu_valid.append(user)
                    yp_valid.append(task)
                elif (seance - 1) % 6 == self.seance_split[2][0]:
                    X_test.append(x)
                    yu_test.append(user)
                    yp_test.append(task)
        return (
            X_train,
            yu_train,
            yp_train,
            X_valid,
            yu_valid,
            yp_valid,
            X_test,
            yu_test,
            yp_test,
        )

    def tensorise(self, X, yu, yp):
        return (
            tensor(np.array(X)).float(),
            one_hot(tensor(np.array(yu))).float(),
            one_hot(tensor(np.array(yp))).float(),
        )


if __name__ == "__main__":
    opp = Opportunity()
    train, valid, test = opp.get_lstm_data()
