import os
from collections import Counter
from datetime import datetime
from random import sample
from warnings import filterwarnings

import numpy as np
import pandas
from pandas import DataFrame, read_csv, read_pickle
from pandas.core.common import SettingWithCopyWarning
from scipy.interpolate import interp1d
from sklearn.model_selection import train_test_split
from torch import tensor
from torch.nn.functional import one_hot
from plotly import graph_objects as go

import settings
from data.cabackend import CabackendDataset

filterwarnings(action="ignore", category=SettingWithCopyWarning)


class Wesad:
    def __init__(
        self,
        data_dir="wesad/",
        verbose=0,
        seed=42069,
        test=False,
        userlim=None,
        preloaded=False,
    ):
        self.data_dir = settings.DATA_DIR + "wesad/"
        self.verbose = verbose
        self.users = sorted(os.listdir(self.data_dir))
        self.data_cols = ["rx", "ry", "rz", "ex", "ey", "ez"]
        self.lookback = settings.LOOKBACK
        self.seed = seed
        # This is used when testing to speed up the import process.
        self.test = test
        self.preloaded = preloaded
        if userlim:
            self.users = sample(self.users, userlim)
            print("Taking a user subset: {}".format(self.users))

    def _load_preloaded(self):

        X_train = np.load(settings.DATA_DIR + "wesad/X_train.npy")
        y_train_user = np.load(settings.DATA_DIR + "wesad/y_train_user.npy")
        y_train_task = np.load(settings.DATA_DIR + "wesad/y_train_task.npy")

        X_valid = np.load(settings.DATA_DIR + "wesad/X_valid.npy")
        y_valid_user = np.load(settings.DATA_DIR + "wesad/y_valid_user.npy")
        y_valid_task = np.load(settings.DATA_DIR + "wesad/y_valid_task.npy")

        X_test = np.load(settings.DATA_DIR + "wesad/X_test.npy")
        y_test_user = np.load(settings.DATA_DIR + "wesad/y_test_user.npy")
        y_test_task = np.load(settings.DATA_DIR + "wesad/y_test_task.npy")

        return (
            CabackendDataset(
                X_train,
                y_train_user,
                y_train_task,
            ),
            CabackendDataset(
                X_valid,
                y_valid_user,
                y_valid_task,
            ),
            CabackendDataset(
                X_test,
                y_test_user,
                y_test_task,
            ),
        )

    @staticmethod
    def interpolate_signal(signal, sampling_rate, base_sampling_rate, label=False):
        # This value should match for signals of different base_sampling_rate
        duration = len(signal) / base_sampling_rate
        if not duration.is_integer():
            raise ValueError("Duration value is not an integer.")
        else:
            duration = int(duration)

        x_base = np.linspace(0, duration, num=len(signal))
        x = np.linspace(0, duration, num=duration * sampling_rate)
        if label:
            interpolate_function = interp1d(x_base, signal, kind="nearest")
        else:
            interpolate_function = interp1d(x_base, signal)

        y = interpolate_function(x)
        return y

    def import_data_test(self):
        """
        Get accelerometer data from both, the Empatica, and RespiBAN devices.
        """
        rx, ry, rz = [], [], []
        ex, ey, ez = [], [], []
        uc, tc = [], []
        sub = 1000
        for user in self.users:
            user_data = read_pickle(self.data_dir + user + "/" + user + ".pkl")
            rx.extend(
                self.interpolate_signal(
                    user_data["signal"]["chest"]["ACC"].T[0], 20, 700
                )[:sub]
            )
            ry.extend(
                self.interpolate_signal(
                    user_data["signal"]["chest"]["ACC"].T[1], 20, 700
                )[:sub]
            )
            rz.extend(
                self.interpolate_signal(
                    user_data["signal"]["chest"]["ACC"].T[2], 20, 700
                )[:sub]
            )
            ex.extend(
                self.interpolate_signal(
                    user_data["signal"]["wrist"]["ACC"].T[0], 20, 32
                )[:sub]
            )
            ey.extend(
                self.interpolate_signal(
                    user_data["signal"]["wrist"]["ACC"].T[1], 20, 32
                )[:sub]
            )
            ez.extend(
                self.interpolate_signal(
                    user_data["signal"]["wrist"]["ACC"].T[2], 20, 32
                )[:sub]
            )
            task_label = self.interpolate_signal(user_data["label"], 20, 700)[:sub]
            uc.extend([int(user[1:]) for _ in range(len(task_label))])
            tc.extend([int(x) for x in task_label])

        lengths = [len(x) for x in [rx, ry, rz, ex, ey, ez, uc, tc]]
        if len(np.unique(lengths)) > 1:
            raise ValueError("Mismatched signal lengths: {}.".format(lengths))

        return {
            "user": uc,
            "task": tc,
            "rx": rx,
            "ry": ry,
            "rz": rz,
            "ex": ex,
            "ey": ey,
            "ez": ez,
        }

    def import_data(self):
        """
        Get accelerometer data from both, the Empatica, and RespiBAN devices.
        """
        rx, ry, rz = [], [], []
        ex, ey, ez = [], [], []
        uc, tc = [], []
        for user in self.users:
            user_data = read_pickle(self.data_dir + user + "/" + user + ".pkl")
            rx.extend(
                self.interpolate_signal(
                    user_data["signal"]["chest"]["ACC"].T[0], 20, 700
                )
            )
            ry.extend(
                self.interpolate_signal(
                    user_data["signal"]["chest"]["ACC"].T[1], 20, 700
                )
            )
            rz.extend(
                self.interpolate_signal(
                    user_data["signal"]["chest"]["ACC"].T[2], 20, 700
                )
            )
            ex.extend(
                self.interpolate_signal(
                    user_data["signal"]["wrist"]["ACC"].T[0], 20, 32
                )
            )
            ey.extend(
                self.interpolate_signal(
                    user_data["signal"]["wrist"]["ACC"].T[1], 20, 32
                )
            )
            ez.extend(
                self.interpolate_signal(
                    user_data["signal"]["wrist"]["ACC"].T[2], 20, 32
                )
            )
            task_label = self.interpolate_signal(user_data["label"], 20, 700)
            uc.extend([int(user[1:]) for _ in range(len(task_label))])
            tc.extend([int(x) for x in task_label])

        lengths = [len(x) for x in [rx, ry, rz, ex, ey, ez, uc, tc]]
        if len(np.unique(lengths)) > 1:
            raise ValueError("Mismatched signal lengths: {}.".format(lengths))

        return {
            "user": uc,
            "task": tc,
            "rx": rx,
            "ry": ry,
            "rz": rz,
            "ex": ex,
            "ey": ey,
            "ez": ez,
        }

    def transform_lstm(self, data):
        results = {
            "rx": [],
            "ry": [],
            "rz": [],
            "ex": [],
            "ey": [],
            "ez": [],
            "user": [],
            "task": [],
        }
        for i in range(self.lookback, len(data["user"])):
            # We also exclude task labels 5, 6, and 7 as they are not relevant for our experiments.
            if (
                len(np.unique(data["user"][i - self.lookback : i])) == 1
                and len(np.unique(data["task"][i - self.lookback : i])) == 1
                and data["task"][i] < 5
                and data["task"][i] > 0
            ):
                for column in self.data_cols:
                    results[column].append(data[column][i - self.lookback : i])
                results["user"].append(data["user"][i])
                results["task"].append(data["task"][i])
        return DataFrame(results)

    def normalise(self, data):
        X = []
        for column in self.data_cols:
            # We have to do this as we store lists in cells of a Pandas
            # DataFrame and Numpy doesn't pick up all array dimensions properly.
            col_data = np.array([y for y in data[column].values])
            col_mean = np.mean(col_data)
            col_std = np.std(col_data)
            norm_data = (col_data - col_mean) / col_std
            X.append(norm_data)
        return np.moveaxis(X, 0, -1)

    def one_hot_encode(self, data, column):
        idx = sorted(np.unique(data[column].values))
        return one_hot(tensor([idx.index(x) for x in data[column].values]))

    def get_lstm_data(self):
        """
        Generate data of shape (N, L, H_in) where;
        N -> batch size
        L -> sequence length (lookback)
        H_in -> input size (number of features
        """
        if self.preloaded:
            return self._load_preloaded()
        s = datetime.now()
        # Import data
        if not self.test:
            data = self.import_data()
        else:
            data = self.import_data_test()

        # Generate LSTM input (lookback)
        data = self.transform_lstm(data)

        # Train test split
        train, test = train_test_split(data, random_state=self.seed, train_size=0.8)
        valid, test = train_test_split(test, random_state=self.seed, train_size=0.5)

        # Normalise input data
        X_train = self.normalise(train)
        X_valid = self.normalise(valid)
        X_test = self.normalise(test)

        # One hot encode
        y_train_user = self.one_hot_encode(train, "user")
        y_train_task = self.one_hot_encode(train, "task")
        y_valid_user = self.one_hot_encode(valid, "user")
        y_valid_task = self.one_hot_encode(valid, "task")
        y_test_user = self.one_hot_encode(test, "user")
        y_test_task = self.one_hot_encode(test, "task")

        # Report on data size
        subtrain = 5 * 2048
        subtest = 2048
        print("Data imported and processed.")
        print("Running time: {}".format(datetime.now() - s))
        print(
            "Total number of rows: {}, {}, {}".format(
                X_train[:subtrain].shape,
                X_valid[:subtest].shape,
                X_test[:subtest].shape,
            )
        )
        print()
        print("User counters:")
        print(Counter(train["user"][:subtrain]))
        print(Counter(valid["user"][:subtest]))
        print(Counter(test["user"][:subtest]))
        print()
        print("Task counters:")
        print(Counter(train["task"][:subtrain]))
        print(Counter(valid["task"][:subtest]))
        print(Counter(test["task"][:subtest]))
        print()
        print("Data shape:")
        print(
            X_train[:subtrain].shape,
            y_train_user[:subtrain].shape,
            y_train_task[:subtrain].shape,
        )
        print(
            X_valid[:subtest].shape,
            y_valid_user[:subtest].shape,
            y_valid_task[:subtest].shape,
        )
        print(
            X_test[:subtest].shape,
            y_test_user[:subtest].shape,
            y_test_task[:subtest].shape,
        )

        # np.save("wesad/X_train", X_train[:subtrain])
        # np.save("wesad/y_train_user", y_train_user[:subtrain])
        # np.save("wesad/y_train_task", y_train_task[:subtrain])
        #
        # np.save("wesad/X_valid", X_valid[:subtest])
        # np.save("wesad/y_valid_user", y_valid_user[:subtest])
        # np.save("wesad/y_valid_task", y_valid_task[:subtest])
        #
        # np.save("wesad/X_test", X_test[:subtest])
        # np.save("wesad/y_test_user", y_test_user[:subtest])
        # np.save("wesad/y_test_task", y_test_task[:subtest])

        # Return torch datasets
        return (
            CabackendDataset(
                X_train[:subtrain],
                y_train_user[:subtrain],
                y_train_task[:subtrain],
            ),
            CabackendDataset(
                X_valid[:subtest],
                y_valid_user[:subtest],
                y_valid_task[:subtest],
            ),
            CabackendDataset(
                X_test[:subtest],
                y_test_user[:subtest],
                y_test_task[:subtest],
            ),
        )


if __name__ == "__main__":
    wesad = Wesad()
    train, valid, test = wesad.get_lstm_data()
