from datetime import datetime
import os
from random import randint

import numpy as np
from keras.layers import BatchNormalization
from tensorflow.python.keras.losses import CategoricalCrossentropy, MeanSquaredError

from data.opportunity import Opportunity
from data.wesad import Wesad
from helpers.soc_comm import ClientBuffer

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"
import wandb
import tensorflow as tf
from sklearn.metrics import accuracy_score, confusion_matrix
from tensorflow.python.keras import Input, Model
from tensorflow.python.keras.layers import (
    Dense,
    Dropout,
    LSTM,
    Conv1D,
    Conv2D,
    MaxPool1D,
    Flatten,
)
from tensorflow.python.keras.optimizer_v2.adam import Adam
from tensorflow.python.ops.math_ops import argmax, reduce_euclidean_norm
from tensorflow.python.ops.numpy_ops.np_arrays import convert_to_tensor
from torch.utils.data import DataLoader

from data.cabackend import Cabackend

# DROPOUT = 0.1
LEARNING_RATE = 0.0001
# LEARNING_RATE = 1
# BATCH_SIZE = 8192
BATCH_SIZE = 2048
EPOCHS = 500
# EPOCHS = 100
EPCH_MULT = 3
DROPOUT = 0.3

NEGATIVE_WEIGHT = -1
POSITIVE_WEIGHT = 0.1

# DATA_SEED = randint(1, 100000)
DATA_SEED = 55796

# How many training epochs before we move to cumulative training (k from PAN)
K = 10
# K = 1

# METHOD = "pan"
METHOD = "ours"

ENC_LSTM = 128
ENC_DENSE = 128

CLASS_01 = 64
CLASS_02 = 32


class MisterStop:
    def __init__(self):
        self.privacy_accs = []

    def check_stop(self):
        print("Checking accs to stop...")
        if self.privacy_accs[-1] < np.mean(self.privacy_accs) + 0.01:
            print("Privacy classifier not progressing, stopping the training process.")
            return True
        print(
            "Last value ({}) higher than adjusted mean ({})".format(
                self.privacy_accs[-1], np.mean(self.privacy_accs) + 0.01
            )
        )
        print("Continuing...")
        return False

    def add_acc(self, privacy_acc):
        self.privacy_accs.append(privacy_acc)


class PlsWorkModelCNN:
    def __init__(self, trainset, validset, testset):
        self.trainset = trainset
        self.validset = validset
        self.testset = testset
        self.run_times = []

        self.i = 0

        # disable_eager_execution()
        # encoder_in = Input(shape=(7, 87), name="enc_in")
        # encoder_in = Input(shape=(16, 36), name="enc_in")
        encoder_in = Input(shape=(7, 6), name="enc_in")
        lstm = LSTM(
            ENC_LSTM,
            # dropout=DROPOUT,
            name="enc_01",
        )(encoder_in)
        drp_out = Dense(ENC_DENSE, activation="relu", name="enc_out")(lstm)

        self.optimizer = Adam(learning_rate=LEARNING_RATE)
        self.loss_fn = CategoricalCrossentropy()

        self.model_encoder = Model(inputs=encoder_in, outputs=drp_out)

        self.stop_check = None
        self.buffer = ClientBuffer()

    def train(self):
        self.model_encoder.summary()
        adv_epocs = EPOCHS * EPCH_MULT

        # Original PAN
        if METHOD == "pan":
            # Not utilised here, but an easy fix for missing mister stop in reporting.
            self.stop_check = MisterStop()
            self.pan_step()
        # Our PAN
        elif METHOD == "ours":
            for i in range(8):
                print("Loop #{}".format(i + 1))
                self.normal_step()
                self.stop_check = MisterStop()
                self.adversary_step(epochs=adv_epocs)
                if self.stop_check.check_stop():
                    break
        else:
            raise ValueError("Invalid method selected.")
        self.adversary_step(epochs=5000)

    def normal_step(self):
        print("NORMAL MODE")
        # Train the network as specified, batch by batch
        for e in range(EPOCHS):
            self.i += 1
            s = datetime.now()

            X_prev = None
            for X, y in self.trainset:
                X = convert_to_tensor(X.numpy())
                y_user = convert_to_tensor(y[0])
                y_task = convert_to_tensor(y[1])

                self.step_01(X, X_prev, y_user)
                self.step_02(X, y_task)
                X_prev = X

            X_prev = None
            for X, y in self.trainset:
                X = convert_to_tensor(X.numpy())
                y_user = convert_to_tensor(y[0])
                y_task = convert_to_tensor(y[1])

                self.step_03(X, X_prev, y_user, y_task)
                X_prev = X

            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))

    def pan_step(self):
        print("PAN MODE")
        # Train the network as specified, batch by batch
        for e in range(EPOCHS):
            self.i += 1
            s = datetime.now()
            for k in range(K):
                for X, y in self.trainset:
                    X = convert_to_tensor(X.numpy())
                    y_user = convert_to_tensor(y[0])
                    y_task = convert_to_tensor(y[1])

                    self.step_01(X, y_user)
                    self.step_02(X, y_task)
                    # The privacy reconstructor step
                    self.step_02_rec(X)

                for X, y in self.trainset:
                    X = convert_to_tensor(X.numpy())
                    y_user = convert_to_tensor(y[0])
                    y_task = convert_to_tensor(y[1])

                    self.step_03(X, y_user, y_task)

            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))
            self.report(
                epoch_time,
                True if e % 100 == 0 else False,
                True if e + 1 == EPOCHS else False,
            )

    def adversary_step(self, epochs=EPOCHS):
        print("ADVERSARY MODE")
        # Only do this when retraining the adversary classifier
        for e in range(epochs):
            self.i += 1
            s = datetime.now()

            for X, y in self.trainset:
                X = convert_to_tensor(X.numpy())
                y_task = convert_to_tensor(y[1])

                self.step_02(X, y_task)

            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))
            self.report(
                epoch_time,
                True if e % 100 == 0 else False,
                True if e + 1 == epochs else False,
                stop=True,
            )

    def step_01(self, X, X_prev, y_user):
        """
        Training of the good classifier.
        """
        logits = self.model_encoder(X)
        print("Sending forward.")
        self.buffer.send_forward(logits)
        self.buffer.send_forward(y_user)
        # In the first epoch we only send the data and
        # wait for the next epoch to perform backpropagation.
        if X_prev == None:
            print("First epoch, no backprop.")
            return

        server_grads = self.buffer.take_backward()
        with tf.GradientTape() as tape:
            logits = self.model_encoder(X_prev, training=True)

        grads = tape.gradient(loss_value, self.model_encoder.trainable_weights)
        print("Received grads.")
        self.optimizer.apply_gradients(zip(grads, self.model_encoder.trainable_weights))

    def step_02(self, X, y_task):
        """
        Training of the bad classifier.
        The encoder does not get updated with feedback from the bad classifier,
        so this step is as simple as calculating logits and passing them to
        the server, alongside with the task labels.
        """
        logits = self.model_encoder(X)
        self.buffer.send_forward(logits)
        self.buffer.send_forward(y_task)

    def step_03(self, X, X_prev, y_user, y_task):
        """
        Training of the whole system with weighted loss functions (on the server side).
        """
        logits = self.model_encoder(X)
        self.buffer.send_forward(logits)
        self.buffer.send_forward(y_user)
        self.buffer.send_forward(y_task)
        if X_prev == None:
            return

        good_grads = self.buffer.take_backward()
        with tf.GradientTape() as tape:
            logits = self.model_encoder(X_prev, training=True)
            loss_value = self.loss_fn(good_grads, logits)
        grads = tape.gradient(loss_value, self.model_encoder.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.model_encoder.trainable_weights))

        bad_grads = self.buffer.take_backward()
        with tf.GradientTape as tape:
            logits = self.model_encoder(X_prev, training=True)
            loss_value = self.loss_fn(bad_grads, logits)
        grads = tape.gradient(loss_value, self.model_encoder.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.model_encoder.trainable_weights))


if __name__ == "__main__":
    gpus = tf.config.experimental.list_physical_devices("GPU")
    if gpus:
        try:
            print(gpus)
            tf.config.experimental.set_virtual_device_configuration(
                gpus[0],
                [tf.config.experimental.VirtualDeviceConfiguration(memory_limit=4096)],
            )
        except RuntimeError as e:
            print(e)

    wesad = Wesad(
        data_dir="/home/andraz/Projects/ppontt_v2.1/data/wesad/",
        seed=DATA_SEED,
        test=False,
    )
    train_data, validation_data, test_data = wesad.get_lstm_data()

    train_loader = DataLoader(
        train_data, batch_size=BATCH_SIZE, shuffle=True, drop_last=False
    )
    valid_loader = DataLoader(
        validation_data, batch_size=BATCH_SIZE, shuffle=False, drop_last=False
    )
    test_loader = DataLoader(
        test_data, batch_size=BATCH_SIZE, shuffle=False, drop_last=False
    )

    model = PlsWorkModelCNN(train_loader, valid_loader, test_loader)
    model.train()
