from datetime import datetime
import os
from random import randint

import numpy as np
from keras.layers import BatchNormalization
from tensorflow.python.keras.losses import CategoricalCrossentropy, MeanSquaredError

import settings
from data.opportunity import Opportunity
from data.wesad import Wesad
from helpers.soc_comm import ServerBuffer

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"
import wandb
import tensorflow as tf
from sklearn.metrics import accuracy_score, confusion_matrix
from tensorflow.python.keras import Input, Model
from tensorflow.python.keras.layers import (
    Dense,
    Dropout,
    LSTM,
    Conv1D,
    Conv2D,
    MaxPool1D,
    Flatten,
)
from tensorflow.python.keras.optimizer_v2.adam import Adam
from tensorflow.python.ops.math_ops import argmax, reduce_euclidean_norm
from tensorflow.python.ops.numpy_ops.np_arrays import convert_to_tensor
from torch.utils.data import DataLoader

from data.cabackend import Cabackend

# DROPOUT = 0.1
LEARNING_RATE = 0.0001
# LEARNING_RATE = 1
# BATCH_SIZE = 8192
BATCH_SIZE = 2048
EPOCHS = 500
# EPOCHS = 100
EPCH_MULT = 3
DROPOUT = 0.3

NEGATIVE_WEIGHT = -1
POSITIVE_WEIGHT = 0.1

# DATA_SEED = randint(1, 100000)
DATA_SEED = 55796

# How many training epochs before we move to cumulative training (k from PAN)
K = 10
# K = 1

# METHOD = "pan"
METHOD = "ours"

ENC_LSTM = 128
ENC_DENSE = 128

CLASS_01 = 64
CLASS_02 = 32


class MisterStop:
    def __init__(self):
        self.privacy_accs = []

    def check_stop(self):
        print("Checking accs to stop...")
        if self.privacy_accs[-1] < np.mean(self.privacy_accs) + 0.01:
            print("Privacy classifier not progressing, stopping the training process.")
            return True
        print(
            "Last value ({}) higher than adjusted mean ({})".format(
                self.privacy_accs[-1], np.mean(self.privacy_accs) + 0.01
            )
        )
        print("Continuing...")
        return False

    def add_acc(self, privacy_acc):
        self.privacy_accs.append(privacy_acc)


class PlsWorkModelCNN:
    def __init__(self):
        self.run_times = []

        self.i = 0

        # disable_eager_execution()

        good_in = Input(shape=ENC_DENSE, name="good_in")
        good_01 = Dense(CLASS_01, activation="relu", name="good_01")(good_in)
        good_drp_01 = Dropout(DROPOUT)(good_01)
        good_02 = Dense(CLASS_02, activation="relu", name="good_02")(good_drp_01)
        good_drp_02 = Dropout(DROPOUT)(good_02)
        good_out = Dense(15, activation="softmax", name="good_classifier")(good_drp_02)

        bad_in = Input(shape=ENC_DENSE, name="bad_in")
        bad_01 = Dense(CLASS_01, activation="relu", name="bad_01")(bad_in)
        bad_drp_01 = Dropout(DROPOUT)(bad_01)
        bad_02 = Dense(CLASS_02, activation="relu", name="bad_02")(bad_drp_01)
        bad_drp_02 = Dropout(DROPOUT)(bad_02)
        bad_out = Dense(4, activation="softmax", name="bad_classifier")(bad_drp_02)

        self.optimizer = Adam(learning_rate=LEARNING_RATE)
        self.loss_fn = CategoricalCrossentropy()

        self.model_good = Model(inputs=good_in, outputs=good_out)
        self.model_bad = Model(inputs=bad_in, outputs=bad_out)

        self.stop_check = None
        self.buffer = ServerBuffer()

        training_params = {
            "Learning Rate": LEARNING_RATE,
            "Batch Size": BATCH_SIZE,
            "Epochs": EPOCHS,
            "Dropout": DROPOUT,
            "Users": settings.CABACKEND_USERS,
            "Positive weight": POSITIVE_WEIGHT,
            "Negative weight": NEGATIVE_WEIGHT,
            "Data seed": DATA_SEED,
            "Encoder lstm size": ENC_LSTM,
            "Encoder dense size": ENC_DENSE,
            "Dense layer 1 size": CLASS_01,
            "Dense layer 2 size": CLASS_02,
            "Method": METHOD,
        }
        # wandb.init(project="ppontt", entity="colosal", config=training_params)

    def train(self):
        self.model_good.summary()
        self.model_bad.summary()
        adv_epocs = EPOCHS * EPCH_MULT

        # Original PAN
        if METHOD == "pan":
            pass
        # Our PAN
        elif METHOD == "ours":
            for i in range(8):
                print("Loop #{}".format(i + 1))
                self.normal_step()
                self.stop_check = MisterStop()
                self.adversary_step(epochs=adv_epocs)
                if self.stop_check.check_stop():
                    break
        else:
            raise ValueError("Invalid method selected.")

        self.adversary_step(epochs=5000)
        # wandb.finish()

    def normal_step(self):
        print("NORMAL MODE")
        # Train the network as specified, batch by batch
        for e in range(EPOCHS):
            self.i += 1
            s = datetime.now()

            while True:
                self.step_01()
                self.step_02()
                break

            while True:
                self.step_03()
                break

            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))
            self.report(
                epoch_time,
                True if e % 100 == 0 else False,
                True if e + 1 == EPOCHS else False,
            )

    def adversary_step(self, epochs=EPOCHS):
        print("ADVERSARY MODE")
        # Only do this when retraining the adversary classifier
        for e in range(epochs):
            self.i += 1
            s = datetime.now()

            for X, y in self.trainset:
                X = convert_to_tensor(X.numpy())
                y_task = convert_to_tensor(y[1])

                self.step_02(X, y_task)

            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))
            self.report(
                epoch_time,
                True if e % 100 == 0 else False,
                True if e + 1 == epochs else False,
                stop=True,
            )

    def step_01(self):
        """
        Training of the good classifier.
        """
        X = self.buffer.take_forward()
        y_user = self.buffer.take_forward()
        print("Received forward.")
        with tf.GradientTape() as tape:
            logits = self.model_good(X, training=True)
            loss_value = self.loss_fn(y_user, logits)

        grads = tape.gradient(loss_value, self.model_good.trainable_weights)
        print("Sending grads.")
        self.buffer.send_backward(grads)
        self.optimizer.apply_gradients(zip(grads, self.model_good.trainable_weights))

    def step_02(self):
        """
        Training of the bad classifier.
        """
        X = self.buffer.take_forward()
        y_task = self.buffer.take_forward()
        with tf.GradientTape() as tape:
            logits = self.model_bad(X, training=True)
            loss_value = self.loss_fn(y_task, logits)
        grads = tape.gradient(loss_value, self.model_bad.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.model_bad.trainable_weights))

    def step_03(self):
        """
        Training of the complete system with negative weight for bad loss.
        """
        X = self.buffer.take_forward()
        y_user = self.buffer.take_forward()
        y_task = self.buffer.take_forward()

        with tf.GradientTape(persistent=True) as tape:
            gl = self.model_good(X)
            loss_good = self.loss_fn(y_user, gl) * POSITIVE_WEIGHT
            bl = self.model_bad(X)
            loss_bad = self.loss_fn(y_task, bl) * NEGATIVE_WEIGHT

        grads = tape.gradient(loss_good, self.model_good.trainable_weights)
        self.buffer.send_backward(grads)
        self.optimizer.apply_gradients(zip(grads, self.model_good.trainable_weights))

        # We do not apply these grads to the bad model, as the loss function is negatively
        # weighted and would artificially hurt the performance of the bad classifier.
        grads = tape.gradient(loss_bad, self.model_bad.trainable_weights)
        self.buffer.send_backward(grads)

    def report(self, epoch_time, cm=False, test=False, stop=False):
        self.run_times.append(epoch_time)

        evaluation = {}
        ul, tl, rl, ua, ta, ucm, tcm = self.evaluate(self.trainset, cm)
        evaluation["Utility/Loss/train"] = ul
        evaluation["Privacy/Loss/train"] = tl
        evaluation["Reconstruction/Loss/train"] = rl
        evaluation["Utility/Accuracy/train"] = ua
        evaluation["Privacy/Accuracy/train"] = ta
        if cm:
            evaluation["Utility/ConfusionMatrix/train"] = ucm
            evaluation["Privacy/ConfusionMatrix/train"] = tcm

        ul, tl, rl, ua, ta, ucm, tcm = self.evaluate(self.validset, cm)
        if stop:
            self.stop_check.add_acc(ta)
        evaluation["Utility/Loss/valid"] = ul
        evaluation["Privacy/Loss/valid"] = tl
        evaluation["Reconstruction/Loss/valid"] = rl
        evaluation["Utility/Accuracy/valid"] = ua
        evaluation["Privacy/Accuracy/valid"] = ta
        if cm:
            evaluation["Utility/ConfusionMatrix/valid"] = ucm
            evaluation["Privacy/ConfusionMatrix/valid"] = tcm

        if test:
            ul, tl, rl, ua, ta, ucm, tcm = self.evaluate(self.testset, False)
            evaluation["Utility/Loss/test"] = ul
            evaluation["Privacy/Loss/test"] = tl
            evaluation["Reconstruction/Loss/test"] = rl
            evaluation["Utility/Accuracy/test"] = ua
            evaluation["Privacy/Accuracy/test"] = ta

            evaluation["Epoch time"] = np.mean(
                [x.total_seconds() for x in self.run_times]
            )

        wandb.log(evaluation)

    def evaluate(self, dataset, cm):
        user_accs = []
        task_accs = []
        user_loss = []
        task_loss = []
        rec_loss = []
        if cm:
            trues, preds, ttrues, tpreds = [], [], [], []
        n = 0
        for X, y in dataset:
            X = convert_to_tensor(X.numpy())
            y_user = convert_to_tensor(y[0])
            y_task = convert_to_tensor(y[1])
            n += len(X)

            y_pred_user = self.model_good(self.model_encoder(X))
            y_pred_task = self.model_bad(self.model_encoder(X))
            # y_pred_rec = self.model_rec(self.model_encoder(X))
            user_loss.append(self.loss_fn(y_user, y_pred_user) * len(X))
            task_loss.append(self.loss_fn(y_task, y_pred_task) * len(X))
            # rec_loss.append(self.loss_rec(X[:, -1, :], y_pred_rec) * len(X))

            y_user = argmax(y_user, 1)
            y_task = argmax(y_task, 1)
            y_pred_user = argmax(y_pred_user, 1)
            y_pred_task = argmax(y_pred_task, 1)
            user_accs.append(accuracy_score(y_user, y_pred_user) * len(X))
            task_accs.append(accuracy_score(y_task, y_pred_task) * len(X))
            if cm:
                trues.extend(y_user)
                preds.extend(y_pred_user)
                ttrues.extend(y_task)
                tpreds.extend(y_pred_task)
        if cm:
            trues = [x.numpy() for x in trues]
            preds = [x.numpy() for x in preds]
            ttrues = [x.numpy() for x in ttrues]
            tpreds = [x.numpy() for x in tpreds]

            return (
                sum(user_loss) / n,
                sum(task_loss) / n,
                sum(rec_loss) / n,
                sum(user_accs) / n,
                sum(task_accs) / n,
                wandb.plot.confusion_matrix(
                    y_true=trues,
                    preds=preds,
                    # class_names=[str(x) for x in settings.CABACKEND_USERS],
                    class_names=[str(x) for x in range(15)],
                ),
                wandb.plot.confusion_matrix(
                    y_true=ttrues,
                    preds=tpreds,
                    class_names=["0", "1", "2", "3", "4", "5", "6", "7"],
                    # , class_names=["1", "2", "3"]
                ),
            )
        else:
            return (
                sum(user_loss) / n,
                sum(task_loss) / n,
                sum(rec_loss) / n,
                sum(user_accs) / n,
                sum(task_accs) / n,
                None,
                None,
            )


if __name__ == "__main__":
    gpus = tf.config.experimental.list_physical_devices("GPU")
    if gpus:
        try:
            print(gpus)
            tf.config.experimental.set_virtual_device_configuration(
                gpus[0],
                [tf.config.experimental.VirtualDeviceConfiguration(memory_limit=4096)],
            )
        except RuntimeError as e:
            print(e)
    model = PlsWorkModelCNN()
    model.train()
