from datetime import datetime
import os
from random import randint
from time import perf_counter, sleep

import numpy as np
import torch
from torch import nn, flatten

from data.cabackend import Cabackend
from data.wesad import Wesad
from helpers.soc_comm import ClientBuffer

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"
from torch.utils.data import DataLoader

# DROPOUT = 0.1
LEARNING_RATE = 0.0001
# LEARNING_RATE = 1
# BATCH_SIZE = 8192
BATCH_SIZE = 2048
EPOCHS = 11
# EPOCHS = 10
EPCH_MULT = 1
DROPOUT = 0.3

NEGATIVE_WEIGHT = -1
POSITIVE_WEIGHT = 0.1

# DATA_SEED = randint(1, 100000)
DATA_SEED = 55796

# How many training epochs before we move to cumulative training (k from PAN)
K = 10
# K = 1

# METHOD = "pan"
METHOD = "ours"

ENC_LSTM = 128
ENC_DENSE = 128

CLASS_01 = 64
CLASS_02 = 32


class Encoder(nn.Module):
    def __init__(self, input_size=6):
        super(Encoder, self).__init__()
        self.lstm = nn.LSTM(input_size=input_size, hidden_size=ENC_LSTM)
        self.dropout = nn.Dropout(DROPOUT)
        self.linear_01 = nn.Linear(ENC_DENSE * 7, ENC_DENSE)
        # self.batch_norm = nn.BatchNorm1d(h1)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.lstm(x)
        x = flatten(x[0], start_dim=1)
        x = self.dropout(x)
        x = self.linear_01(x)
        # x = self.batch_norm(x)
        x = self.relu(x)
        return x


class PlsWorkModelCNN:
    def __init__(self, trainset, validset, testset):
        self.trainset = trainset
        self.validset = validset
        self.testset = testset
        self.run_times = []
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

        self.i = 0
        # Define model
        self.model_encoder = Encoder().float().to(self.device)
        # Define optimizer
        self.optimizer = torch.optim.Adam(
            self.model_encoder.parameters(), lr=LEARNING_RATE
        )
        self.stop_check = None
        self.buffer = ClientBuffer()

        self.calc_times = []
        self.comm_times = []
        self.agg_calc_times = []
        self.agg_comm_times = []

    def train(self):
        # self.model_encoder.summary()
        adv_epocs = EPOCHS * EPCH_MULT

        # Original PAN
        if METHOD == "pan":
            # Not utilised here, but an easy fix for missing mister stop in reporting.
            self.pan_step()
        # Our PAN
        elif METHOD == "ours":
            # for i in range(8):
            #     print("Loop #{}".format(i + 1))
            self.normal_step()
            print(
                "Calc times: {:.4f},{:.4f}".format(
                    np.mean(self.agg_calc_times[1:]), np.std(self.agg_calc_times[1:])
                )
            )
            print(
                "Comm times: {:.4f},{:.4f}".format(
                    np.mean(self.agg_comm_times[1:]), np.std(self.agg_comm_times[1:])
                )
            )
            # self.adversary_step(epochs=adv_epocs)
            self.buffer.halt()
            sleep(5)
        else:
            raise ValueError("Invalid method selected.")
        # self.buffer.send_forward("MODE_OUT")
        # self.adversary_step(epochs=5000)
        # self.buffer.send_forward("MODE_STOP")
        # print("Training complete.")

    def normal_step(self):
        print("NORMAL MODE")
        # Train the network as specified, batch by batch
        for e in range(EPOCHS):
            self.i += 1
            self.model_encoder.train()

            # Steps 1 & 2
            self.buffer.send_forward("MODE_STEP1")
            X_prev = None
            for X, y in self.trainset:
                X = X.float().to(self.device)
                y_user = y[0].float().to(self.device)
                y_task = y[1].float().to(self.device)
                self.step_01(X, X_prev, y_user)
                self.step_02(X, y_task)
                X_prev = X

            # Step 3
            self.buffer.send_forward("MODE_STEP3")
            X_prev = None
            for X, y in self.trainset:
                X = X.float().to(self.device)
                y_user = y[0].float().to(self.device)
                y_task = y[1].float().to(self.device)
                self.step_03(X, X_prev, y_user, y_task)
                X_prev = X

            # print(
            #     "Epoch times: {:.4f} calc, {:.4f} comm".format(
            #         sum(self.calc_times), sum(self.comm_times)
            #     )
            # )
            self.agg_calc_times.append((sum(self.calc_times), sum(self.calc_times)))
            self.agg_comm_times.append((sum(self.comm_times), sum(self.comm_times)))
            self.calc_times = []
            self.comm_times = []

            # Testing
            # # self.model_encoder.eval()
            self.buffer.send_forward("MODE_TEST_TRAIN")
            # self.buffer.send_forward("MODE_STEP1")
            # # print("Testing on trainset.")
            # for X, y in self.trainset:
            #     X = X.float().to(self.device)
            #     y_user = y[0].float().to(self.device)
            #     y_task = y[1].float().to(self.device)
            #     self.step_test(X, y_user, y_task)
            # self.buffer.send_forward("MODE_TEST_VALID")
            # # print("Testing on validset.")
            # for X, y in self.validset:
            #     X = X.float().to(self.device)
            #     y_user = y[0].float().to(self.device)
            #     y_task = y[1].float().to(self.device)
            #     self.step_test(X, y_user, y_task)
            # print("Testing complete.")

            # epoch_time = datetime.now() - s
            # if e % 100 == 0:
            # print("Epoch ({}) run time: {}".format(self.i, epoch_time))

    def adversary_step(self, epochs=EPOCHS):
        print("ADVERSARY MODE")
        self.model_encoder.eval()
        # Only do this when retraining the adversary classifier
        for e in range(epochs):
            self.buffer.send_forward("MODE_STEPADV")
            torch.cuda.empty_cache()
            self.i += 1
            s = datetime.now()

            for X, y in self.trainset:
                X = X.float().to(self.device)
                y_task = y[1].float().to(self.device)
                self.step_02(X, y_task)

            # Testing
            self.buffer.send_forward("MODE_TEST_TRAIN")
            for X, y in self.trainset:
                X = X.float().to(self.device)
                y_user = y[0].float().to(self.device)
                y_task = y[1].float().to(self.device)
                self.step_test(X, y_user, y_task)
            self.buffer.send_forward("MODE_TEST_VALID")
            # print("Testing on validset.")
            for X, y in self.validset:
                X = X.float().to(self.device)
                y_user = y[0].float().to(self.device)
                y_task = y[1].float().to(self.device)
                self.step_test(X, y_user, y_task)

            epoch_time = datetime.now() - s
            # if e % 100 == 0:
            print("Epoch ({}) run time: {}".format(self.i, epoch_time))

    def step_01(self, X, X_prev, y_user):
        """
        Training of the good classifier.
        """
        tstart = perf_counter()
        logits = self.model_encoder(X)
        self.calc_times.append(perf_counter() - tstart)

        # print("Sending user.")
        tstart = perf_counter()
        self.buffer.send_forward(logits)
        self.buffer.send_forward(y_user)
        self.comm_times.append(perf_counter() - tstart)
        # In the first epoch we only send the data and
        # wait for the next epoch to perform backpropagation.
        if X_prev == None:
            # print("First epoch, no backprop.")
            return

        tstart = perf_counter()
        X_grad = self.buffer.take_backward()
        self.comm_times.append(perf_counter() - tstart)
        # print("Received user grads.")
        X_grad = X_grad.to(self.device)

        tstart = perf_counter()
        X_prev_out = self.model_encoder(X_prev)
        self.optimizer.zero_grad()
        X_prev_out.backward(X_grad)
        self.calc_times.append(perf_counter() - tstart)

    def step_02(self, X, y_task):
        """
        Training of the bad classifier.
        The encoder does not get updated with feedback from the bad classifier,
        so this step is as simple as calculating logits and passing them to
        the server, alongside with the task labels.
        """
        tstart = perf_counter()
        logits = self.model_encoder(X)
        self.calc_times.append(perf_counter() - tstart)

        tstart = perf_counter()
        self.buffer.send_forward(logits)
        self.buffer.send_forward(y_task)
        self.comm_times.append(perf_counter() - tstart)
        # print("Sending task.")

    def step_03(self, X, X_prev, y_user, y_task):
        """
        Training of the whole system with weighted loss functions (on the server side).
        """
        tstart = perf_counter()
        logits = self.model_encoder(X)
        self.calc_times.append(perf_counter() - tstart)
        # print("Sending both.")

        tstart = perf_counter()
        self.buffer.send_forward(logits)
        self.buffer.send_forward(y_user)
        self.buffer.send_forward(y_task)
        self.comm_times.append(perf_counter() - tstart)
        if X_prev == None:
            # print("First epoch, no backprop.")
            # X_grad = self.buffer.take_backward()
            return

        # The first grad we receive here is still from step 1, as server sends out the grad every
        # epoch, while on the client side we skip the first epoch in terms of backpropagation.

        tstart = perf_counter()
        X_grad = self.buffer.take_backward()
        self.comm_times.append(perf_counter() - tstart)
        # print("Received both grads.")
        X_grad = X_grad.to(self.device)

        tstart = perf_counter()
        X_prev_out = self.model_encoder(X_prev)
        self.optimizer.zero_grad()
        X_prev_out.backward(X_grad)
        self.calc_times.append(perf_counter() - tstart)

    def step_test(self, X, y_user, y_task):
        """
        Send data through the encoder for the server to do the tests.
        """
        logits = self.model_encoder(X)
        self.buffer.send_forward(logits)
        self.buffer.send_forward(y_user)
        self.buffer.send_forward(y_task)
        # print("Sending test.")


if __name__ == "__main__":
    for bandwith_limit in range(20, 8, -1):
        os.system("sudo wondershaper -a enp1s0 -c")
        os.system(
            "sudo wondershaper -a enp1s0 -d {} -u {}".format(
                2**bandwith_limit, 2**bandwith_limit
            )
        )
        print()
        print("Bandwith limited to: {} Mbit/s".format(2 ** (bandwith_limit - 10)))
        wesad = Wesad(
            data_dir="/wesad/",
            seed=DATA_SEED,
            test=False,
            preloaded=True,
        )
        train_data, validation_data, test_data = wesad.get_lstm_data()

        train_loader = DataLoader(
            train_data, batch_size=BATCH_SIZE, shuffle=True, drop_last=True
        )
        valid_loader = DataLoader(
            validation_data, batch_size=BATCH_SIZE, shuffle=True, drop_last=True
        )
        test_loader = DataLoader(
            test_data, batch_size=BATCH_SIZE, shuffle=False, drop_last=False
        )

        model = PlsWorkModelCNN(train_loader, valid_loader, test_loader)
        model.train()
