from datetime import datetime
import os

import numpy as np
import torch
from torch import nn, argmax

import settings
from helpers.soc_comm import ServerBuffer

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"
import wandb
from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score

WANDB_GROUP = "noise_0004"
# DROPOUT = 0.1
LEARNING_RATE = 0.0001
BATCH_SIZE = 2048
EPOCHS = 250
EPCH_MULT = 3
DROPOUT = 0.3

NEGATIVE_WEIGHT = -1
POSITIVE_WEIGHT = 0.1

DATA_SEED = 55796
METHOD = "ours_to"

ENC_LSTM = 609
ENC_DENSE = 609

CLASS_01 = 64
CLASS_02 = 32

NOISE_TYPE = "laplace"
# Increase this by 4x every next run
NOISE_SCALE = 32
DATASET = "cabackend"
# DATASET = "wesad"


class MisterStop:
    def __init__(self):
        self.privacy_accs = []

    def check_stop(self):
        print("Checking accs to stop...")
        if self.privacy_accs[-1] < np.mean(self.privacy_accs) + 0.01:
            print("Privacy classifier not progressing, stopping the training process.")
            return True
        print(
            "Last value ({}) higher than adjusted mean ({})".format(
                self.privacy_accs[-1], np.mean(self.privacy_accs) + 0.01
            )
        )
        print("Continuing...")
        return False

    def add_acc(self, privacy_acc):
        self.privacy_accs.append(privacy_acc)


class GoodClassifier(nn.Module):
    def __init__(self):
        super(GoodClassifier, self).__init__()
        self.linear01 = nn.Linear(ENC_DENSE, CLASS_01)
        self.linear02 = nn.Linear(CLASS_01, CLASS_02)
        self.linear03 = nn.Linear(CLASS_02, 15)
        self.relu = nn.ReLU()
        self.softmax = nn.Softmax(dim=1)
        self.dropout = nn.Dropout(DROPOUT)

    def forward(self, x):
        x = self.linear01(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = self.linear02(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = self.linear03(x)
        x = self.softmax(x)
        return x


class BadClassifier(nn.Module):
    def __init__(self):
        super(BadClassifier, self).__init__()
        self.linear01 = nn.Linear(ENC_DENSE, CLASS_01)
        self.linear02 = nn.Linear(CLASS_01, CLASS_02)
        self.linear03 = nn.Linear(CLASS_02, 3)
        self.relu = nn.ReLU()
        self.softmax = nn.Softmax(dim=1)
        self.dropout = nn.Dropout(DROPOUT)

    def forward(self, x):
        x = self.linear01(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = self.linear02(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = self.linear03(x)
        x = self.softmax(x)
        return x


class ToServer:
    def __init__(self):
        self.run_times = []

        self.i = 0

        self.model_good = GoodClassifier().float().to("cuda")
        self.model_bad = BadClassifier().float().to("cuda")

        self.good_optimizer = torch.optim.Adam(
            self.model_good.parameters(), lr=LEARNING_RATE
        )
        self.bad_optimizer = torch.optim.Adam(
            self.model_bad.parameters(), lr=LEARNING_RATE
        )

        self.loss = torch.nn.CrossEntropyLoss()

        self.stop_check = None
        self.buffer = ServerBuffer()
        self.mode_semaphore = None

        training_params = {
            "Learning Rate": LEARNING_RATE,
            "Batch Size": BATCH_SIZE,
            "Epochs": EPOCHS,
            "Dropout": DROPOUT,
            "Users": settings.CABACKEND_USERS,
            "Positive weight": POSITIVE_WEIGHT,
            "Negative weight": NEGATIVE_WEIGHT,
            "Data seed": DATA_SEED,
            "Encoder lstm size": ENC_LSTM,
            "Encoder dense size": ENC_DENSE,
            "Dense layer 1 size": CLASS_01,
            "Dense layer 2 size": CLASS_02,
            "Method": METHOD,
            "Noise type": NOISE_TYPE,
            "Noise scale": NOISE_SCALE,
            "Dataset": DATASET,
        }
        wandb.init(
            project="ppontt",
            entity="colosal",
            config=training_params,
            group=WANDB_GROUP,
        )

    def train(self):
        # Original PAN
        if METHOD == "pan":
            pass
        # Our PAN
        elif METHOD[0:4] == "ours":
            self.mode_semaphore = self.buffer.take_forward()
            i = 1
            while True:
                print("Loop #{}".format(i))
                self.normal_step()
                self.stop_check = MisterStop()
                self.adversary_step()
                if self.stop_check.check_stop():
                    self.buffer.send_backward("MODE_ENDADV")
                    break
                if self.mode_semaphore == "MODE_OUT":
                    self.buffer.send_backward("MODE_ENDADV")
                    break
                i += 1
        else:
            raise ValueError("Invalid method selected.")
        self.buffer.take_forward()
        self.buffer.take_forward()
        self.mode_semaphore = self.buffer.take_forward()
        self.adversary_step()
        # TODO: test if this works as it should
        self.report(test=True)
        wandb.finish()
        print("Training complete.")

    def normal_step(self):
        print("NORMAL MODE")
        e = 0
        while True:
            self.i += 1
            s = datetime.now()
            self.model_good.train()
            self.model_bad.train()

            # Receive step1 or adv

            # Steps 1 & 2
            while self.mode_semaphore == "MODE_STEP1":
                self.step_01()
                if self.mode_semaphore != "MODE_STEP1":
                    break
                self.step_02()
            # print("Moving to step 3.")

            # Step 3
            while self.mode_semaphore == "MODE_STEP3":
                self.step_03()
            # print("Moving to testing.")
            epoch_time = datetime.now() - s

            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))
            self.model_good.eval()
            self.model_bad.eval()
            self.report(
                epoch_time,
                True if e % 100 == 0 else False,
                False,
            )
            e += 1
            if self.mode_semaphore == "MODE_STEPADV":
                return

    def adversary_step(self):
        print("ADVERSARY MODE")
        # Only do this when retraining the adversary classifier
        self.model_good.eval()
        e = 0
        while True:
            self.i += 1
            s = datetime.now()

            self.model_bad.train()
            while self.mode_semaphore == "MODE_STEPADV":
                self.step_02()

            self.model_bad.eval()
            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))
            self.report(
                epoch_time,
                True if e % 100 == 0 else False,
                False,
                stop=True,
            )
            e += 1
            if (
                self.mode_semaphore == "MODE_STEP1"
                or self.mode_semaphore == "MODE_OUT"
                or self.mode_semaphore == "MODE_STOP"
            ):
                return

    def step_01(self):
        """
        Training of the good classifier.
        """
        X = self.buffer.take_forward()
        if X == "MODE_STEP3" or X == "MODE_STEPADV":
            self.mode_semaphore = X
            return
        y_user = self.buffer.take_forward()
        # print("Received user.")
        X = X.to("cuda")
        X.requires_grad = True
        y_user = y_user.to("cuda")
        X.retain_grad()

        y_pred = self.model_good(X)
        loss = self.loss(y_pred, y_user.clone().detach())
        self.good_optimizer.zero_grad()
        loss.backward()
        self.good_optimizer.step()
        # print("Sending user grads.")
        self.buffer.send_backward(X.grad.data.to("cpu"))

    def step_02(self):
        """
        Training of the bad classifier.
        """
        X = self.buffer.take_forward()
        if X == "MODE_TEST_TRAIN":
            self.mode_semaphore = X
            return
        y_task = self.buffer.take_forward()
        # print("Received task.")

        X = X.to("cuda")
        X.requires_grad = True
        y_task = y_task.to("cuda")
        X.retain_grad()

        y_pred = self.model_bad(X)
        loss = self.loss(y_pred, y_task.clone().detach())
        self.bad_optimizer.zero_grad()
        loss.backward()
        self.bad_optimizer.step()

    def step_03(self):
        """
        Training of the complete system with negative weight for bad loss.
        """
        X = self.buffer.take_forward()
        if X == "MODE_TEST_TRAIN":
            self.mode_semaphore = X
            return
        y_user = self.buffer.take_forward()
        y_task = self.buffer.take_forward()
        # print("Received both.")

        X = X.to("cuda")
        X.requires_grad = True
        X.retain_grad()
        y_user = y_user.to("cuda")
        y_task = y_task.to("cuda")

        y_user_pred = self.model_good(X)
        y_task_pred = self.model_bad(X)
        loss = self.loss(
            y_user_pred, y_user.clone().detach()
        ) * POSITIVE_WEIGHT + self.loss(
            y_task_pred, y_task.clone().detach() * NEGATIVE_WEIGHT
        )

        self.good_optimizer.zero_grad()
        loss.backward()
        self.good_optimizer.step()
        # print("Sending both grads.")
        self.buffer.send_backward(X.grad.data.to("cpu"))

    def report(self, epoch_time=None, cm=False, test=False, stop=False):
        if epoch_time:
            self.run_times.append(epoch_time)

        evaluation = {}
        ul, tl, rl, ua, ta, uf, tf, up, tp, ur, tr, ucm, tcm = self.evaluate(cm)
        if sum((ul, tl, rl, ua, ta)) == 0:
            return
        evaluation["Utility/Loss/train"] = ul
        evaluation["Privacy/Loss/train"] = tl
        evaluation["Reconstruction/Loss/train"] = rl
        evaluation["Utility/Accuracy/train"] = ua
        evaluation["Privacy/Accuracy/train"] = ta
        evaluation["Utility/F1score/train"] = uf
        evaluation["Privacy/F1score/train"] = tf
        evaluation["Utility/Precision/train"] = up
        evaluation["Privacy/Precision/train"] = tp
        evaluation["Utility/Recall/train"] = ur
        evaluation["Privacy/Recall/train"] = tr
        if cm:
            evaluation["Utility/ConfusionMatrix/train"] = ucm
            evaluation["Privacy/ConfusionMatrix/train"] = tcm

        ul, tl, rl, ua, ta, uf, tf, up, tp, ur, tr, ucm, tcm = self.evaluate(cm)
        if stop:
            self.stop_check.add_acc(ta)
        evaluation["Utility/Loss/valid"] = ul
        evaluation["Privacy/Loss/valid"] = tl
        evaluation["Reconstruction/Loss/valid"] = rl
        evaluation["Utility/Accuracy/valid"] = ua
        evaluation["Privacy/Accuracy/valid"] = ta
        evaluation["Utility/F1score/valid"] = uf
        evaluation["Privacy/F1score/valid"] = tf
        evaluation["Utility/Precision/valid"] = up
        evaluation["Privacy/Precision/valid"] = tp
        evaluation["Utility/Recall/valid"] = ur
        evaluation["Privacy/Recall/valid"] = tr
        if cm:
            evaluation["Utility/ConfusionMatrix/valid"] = ucm
            evaluation["Privacy/ConfusionMatrix/valid"] = tcm

        if test:
            ul, tl, rl, ua, ta, uf, tf, up, tp, ur, tr, ucm, tcm = self.evaluate(cm)
            evaluation["Utility/Loss/test"] = ul
            evaluation["Privacy/Loss/test"] = tl
            evaluation["Reconstruction/Loss/test"] = rl
            evaluation["Utility/Accuracy/test"] = ua
            evaluation["Privacy/Accuracy/test"] = ta
            evaluation["Utility/F1score/test"] = uf
            evaluation["Privacy/F1score/test"] = tf
            evaluation["Utility/Precision/test"] = up
            evaluation["Privacy/Precision/test"] = tp
            evaluation["Utility/Recall/test"] = ur
            evaluation["Privacy/Recall/test"] = tr

            evaluation["Epoch time"] = np.mean(
                [x.total_seconds() for x in self.run_times]
            )
        wandb.log(evaluation)

    def evaluate(self, cm):
        user_accs = []
        task_accs = []
        user_loss = []
        task_loss = []
        user_f1 = []
        task_f1 = []
        user_prec = []
        task_prec = []
        user_reca = []
        task_reca = []
        rec_loss = []
        if cm:
            trues, preds, ttrues, tpreds = [], [], [], []
        n = 0
        while True:
            X = self.buffer.take_forward()
            if type(X) == str:
                self.mode_semaphore = X
                if n == 0:
                    pass
                break
            y_user = self.buffer.take_forward()
            y_task = self.buffer.take_forward()
            n += len(X)
            X = X.to("cuda")
            y_user = y_user.to("cuda")
            y_task = y_task.to("cuda")

            y_pred_user = self.model_good(X)
            y_pred_task = self.model_bad(X)
            # y_pred_rec = self.model_rec(self.model_encoder(X))
            user_loss.append(
                self.loss(y_user, y_pred_user).cpu().detach().numpy() * len(X)
            )
            task_loss.append(
                self.loss(y_task, y_pred_task).cpu().detach().numpy() * len(X)
            )
            # rec_loss.append(self.loss_rec(X[:, -1, :], y_pred_rec) * len(X))

            y_user = argmax(y_user.cpu().detach(), 1).numpy()
            y_task = argmax(y_task.cpu().detach(), 1).numpy()
            y_pred_user = argmax(y_pred_user.cpu().detach(), 1).numpy()
            y_pred_task = argmax(y_pred_task.cpu().detach(), 1).numpy()
            user_accs.append(accuracy_score(y_user, y_pred_user) * len(X))
            task_accs.append(accuracy_score(y_task, y_pred_task) * len(X))
            user_f1.append(f1_score(y_user, y_pred_user, average="weighted") * len(X))
            task_f1.append(f1_score(y_task, y_pred_task, average="weighted") * len(X))
            user_prec.append(
                precision_score(
                    y_user, y_pred_user, average="weighted", zero_division=0.0
                )
                * len(X)
            )
            task_prec.append(
                precision_score(
                    y_task, y_pred_task, average="weighted", zero_division=0.0
                )
                * len(X)
            )
            user_reca.append(
                recall_score(y_user, y_pred_user, average="weighted") * len(X)
            )
            task_reca.append(
                recall_score(y_task, y_pred_task, average="weighted") * len(X)
            )
            if cm:
                trues.extend(y_user)
                preds.extend(y_pred_user)
                ttrues.extend(y_task)
                tpreds.extend(y_pred_task)
        if n == 0 and X == "MODE_STEPADV":
            return 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
        if cm:
            # trues = [x.numpy() for x in trues]
            # preds = [x.numpy() for x in preds]
            # ttrues = [x.numpy() for x in ttrues]
            # tpreds = [x.numpy() for x in tpreds]

            return (
                sum(user_loss) / n,
                sum(task_loss) / n,
                sum(rec_loss) / n,
                sum(user_accs) / n,
                sum(task_accs) / n,
                sum(user_f1) / n,
                sum(task_f1) / n,
                sum(user_prec) / n,
                sum(task_prec) / n,
                sum(user_reca) / n,
                sum(task_reca) / n,
                None,
                None
                # wandb.plot.confusion_matrix(
                #     y_true=trues,
                #     preds=preds,
                #     # class_names=[str(x) for x in settings.CABACKEND_USERS],
                #     class_names=[str(x) for x in range(15)],
                # ),
                # wandb.plot.confusion_matrix(
                #     y_true=ttrues,
                #     preds=tpreds,
                #     class_names=["0", "1", "2", "3", "4", "5", "6", "7"],
                #     # , class_names=["1", "2", "3"]
                # ),
            )
        else:
            return (
                sum(user_loss) / n,
                sum(task_loss) / n,
                sum(rec_loss) / n,
                sum(user_accs) / n,
                sum(task_accs) / n,
                sum(user_f1) / n,
                sum(task_f1) / n,
                sum(user_prec) / n,
                sum(task_prec) / n,
                sum(user_reca) / n,
                sum(task_reca) / n,
                None,
                None,
            )


if __name__ == "__main__":
    model = ToServer()
    model.train()
