from __future__ import division

import logging
import sys
import threading

import torch
import torch.backends.cudnn as cudnn
from torch import quantization
from torch.utils.data import DataLoader
from torchinfo import summary

import settings
from data.cabackend import Cabackend
from helpers.logger import setup_logger
from helpers.model import ModelHandler
from helpers.soc_comm import ClientBuffer
from models.client import (
    ClientEncoderNetwork,
    ClientEncoderLSTMNetwork,
    ClientEncoderGRUNetwork,
)

from training.client import train_start


def main():
    print("python version : {}".format(sys.version.replace("\n", " ")))
    print("torch  version : {}".format(torch.__version__))
    print("cudnn  version : {}".format(torch.backends.cudnn.version()))
    # device = "cpu"
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("device         : {}".format(device))
    print()
    print("TRAINING PARAMETERS")
    print("Architecture: {}".format(settings.ARCHITECTURE))
    print("Epochs: {}".format(settings.EPOCHS))
    print("Learning rate: {}".format(settings.LEARNING_RATE))
    print("Batch size: {}".format(settings.BATCH_SIZE))
    print("Dropout: {}".format(settings.DROPOUT))
    print("Data timeout: {} seconds".format(settings.TIMEOUT))
    print()

    # Init loggers
    setup_logger("Client", "log/client.log")
    client_logger = logging.getLogger("Client")

    # Init dataset
    cabackend = Cabackend(split=[0.70, 0.15, 0.15])
    train_data, validation_data, _ = cabackend.get_lstm_data()

    train_loader = DataLoader(
        train_data, batch_size=settings.BATCH_SIZE, shuffle=True, drop_last=False
    )
    valid_loader = DataLoader(validation_data, batch_size=settings.BATCH_SIZE)

    # Init model and optimizer
    encoder_model = ClientEncoderLSTMNetwork().float().to(device)
    encoder_optimizer = torch.optim.Adam(
        encoder_model.parameters(),
        lr=settings.LEARNING_RATE,
    )

    encoder_model.to(device)
    summary(encoder_model)

    # Socket communication buffer
    buffer = ClientBuffer()

    # Run the training
    train_start(
        train_loader,
        valid_loader,
        encoder_model,
        encoder_optimizer,
        buffer,
        client_logger,
        device,
    )


if __name__ == "__main__":
    main()
