from time import sleep

import torch
from torch import quantization, argmax
from torch.optim.lr_scheduler import ExponentialLR

import settings
from helpers.boost_sampler import BoostSampler
from helpers.model import ModelHandler

import numpy as np


def train_start(train_loader, valid_loader, model, optimizer, buffer, logger, device):
    """
    Client training function. Following the 3-step PAN training algorithm.
    We send each batch through the encoder three times, updating the weights on the first and third pass.
    """
    model_handler = ModelHandler(server=False)
    # sampler = BoostSampler(train_loader, batch_size=2**12)
    scheduler = ExponentialLR(optimizer, 0.999)
    for epoch in range(settings.EPOCHS):
        model.train()

        # This is with the default sampler and loader. Now we try to implement our own.
        for i, (X_train, y_train) in enumerate(train_loader):

            # Our own sampler and loader
            # X_train, y_train = sampler.get_epoch()
            # i = 0

            logger.info("Epoch {}, Batch {}".format(epoch, i))
            X_train = X_train.to(device)
            y_train[0] = y_train[0].to(device)
            y_train[1] = y_train[1].to(device)

            # On the first batch of the first epoch, we just pass the data through and send it to the server
            # It's the same thing three times, but for sake of clarity, it's all written out (no real penalty in terms of computational cost)
            if epoch == 0 and i == 0:
                # ============ Step 1 ============
                with torch.no_grad():
                    utility_in = model(X_train)
                logger.info("Sending to forward")
                buffer.send_forward(utility_in)
                buffer.send_forward(y_train)

                # ============ Step 2 ============
                with torch.no_grad():
                    privacy_in = model(X_train)

                logger.info("Sending to forward")
                buffer.send_forward(privacy_in)
                buffer.send_forward(y_train)

                # ============ Step 3 ============
                with torch.no_grad():
                    privacy_in = model(X_train)

                logger.info("Sending to forward")
                buffer.send_forward(privacy_in)
                buffer.send_forward(y_train)

                X_prev = X_train
                continue
            else:
                # ============ Step 1 ============
                logger.info("STEP 1")
                with torch.no_grad():
                    utility_in = model(X_train)

                logger.info("Sending to forward")
                buffer.send_forward(utility_in)
                buffer.send_forward(y_train)

                logger.info("Taking gradients")
                X_grad = buffer.take_backward()
                X_grad = X_grad.to(device)
                logger.info("Gradients received")

                X_prev_out = model(X_prev)
                optimizer.zero_grad()
                X_prev_out.backward(X_grad)
                optimizer.step()

                # ============ Step 2 ============
                logger.info("STEP 2")
                with torch.no_grad():
                    X_out = model(X_train)

                logger.info("Sending to forward")
                buffer.send_forward(X_out)
                buffer.send_forward(y_train)

                # ============ Step 3 ============
                logger.info("STEP 3")
                with torch.no_grad():
                    X_out = model(X_train)

                logger.info("Sending to forward")
                buffer.send_forward(X_out)
                buffer.send_forward(y_train)

                logger.info("Taking gradients")
                X_grad = buffer.take_backward()
                X_grad = X_grad.to(device)
                logger.info("Gradients received")

                X_prev_out = model(X_prev)

                optimizer.zero_grad()
                X_prev_out.backward(X_grad)
                optimizer.step()

                # ============ Remember previous input ============
                X_prev = X_train

        # Testing after every epoch
        # (change the 1 if you want testing/reporting only every n epochs)
        scheduler.step()
        if epoch % 1 == 0:
            model.eval()
            logger.warning("Testing")
            buffer.send_forward(["MODE_TRAIN_TEST"])
            for i, (X_test, y_test) in enumerate(train_loader):
                X_test = X_test.to(device)
                y_test[0] = y_test[0].to(device)
                y_test[1] = y_test[1].to(device)
                X_out = model(X_test)
                buffer.send_forward(X_out)
                buffer.send_forward(y_test)

            buffer.send_forward(["MODE_TEST"])
            for i, (X_test, y_test) in enumerate(valid_loader):
                X_test = X_test.to(device)
                y_test[0] = y_test[0].to(device)
                y_test[1] = y_test[1].to(device)
                X_out = model(X_test)
                buffer.send_forward(X_out)
                buffer.send_forward(y_test)

            # Save model
            model_handler.save_model("encoder_model.pt", model, optimizer, epoch, 0, 0)
            buffer.send_forward(["MODE_TRAIN"])

        else:
            buffer.send_forward(["MODE_TRAIN_TEST"])
            buffer.send_forward(["MODE_TEST"])
            buffer.send_forward(["MODE_TRAIN"])

    logger.warning("Client training completed.")
