from datetime import datetime
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import seaborn as sn

import torch
import wandb as wandb
from numpy import array, mean, std
from sklearn.metrics import confusion_matrix
from torch import argmax, tensor
from torch.nn import L1Loss
from torch.optim.lr_scheduler import ExponentialLR
from torch.utils.tensorboard import SummaryWriter
from torchinfo import summary

import settings
from helpers.model import ModelHandler


def generate_confusion_matrix(y_true, y_pred):
    # https://christianbernecker.medium.com/how-to-create-a-confusion-matrix-with-tensorboard-and-pytorch-3344ad5e7209
    # Check out this as well: https://towardsdatascience.com/a-complete-guide-to-using-tensorboard-with-pytorch-53cb2301e8c3
    # constant for classes
    classes = settings.CABACKEND_USERS

    # Build confusion matrix
    cf_matrix = confusion_matrix(y_true, y_pred)
    df_cm = pd.DataFrame(
        cf_matrix / np.sum(cf_matrix) * 10,
        index=[i for i in classes],
        columns=[i for i in classes],
    )
    plt.figure(figsize=(12, 7))
    return sn.heatmap(df_cm, annot=True).get_figure()


def train_end(models, loss_fn, optimizers, buffer, logger, device):
    """
    Server training function. The three training steps:
    - training of the utility model (update weights of the utility model and send back to encoder)
    - training of the privacy model (update weights of the privacy model)
    - training of the system (update utility model and send back a weighted loss function)
    """

    print("Server training started")
    utility_model, privacy_model = models
    utility_optimizer, privacy_optimizer = optimizers
    utility_scheduler = ExponentialLR(utility_optimizer, 0.999)
    privacy_scheduler = ExponentialLR(privacy_optimizer, 0.999)

    training_params = {
        "Learning Rate": settings.LEARNING_RATE,
        "Batch Size": settings.BATCH_SIZE,
        "Architecture": settings.ARCHITECTURE,
        "Epochs": settings.EPOCHS,
    }
    if settings.ARCHITECTURE == "lstm" or settings.ARCHITECTURE == "gru":
        training_params |= {"Lookback": settings.LOOKBACK}

    wandb.init(project="ppontt", entity="colosal", config=training_params)

    running_times = []

    for epoch in range(settings.EPOCHS):
        utility_running_loss = 0
        privacy_running_loss = 0
        count_running_loss = 0
        utility_model.train()
        privacy_model.train()
        start_time = datetime.now()

        while True:
            count_running_loss += 1
            # ============ Step 1 ============
            logger.info("STEP 1")
            logger.info("Taking forward")
            X_train = buffer.take_forward()

            # The server has no idea about batch sizes, so we trigger
            # the testing mode by sending a MODE_TEST from the client.
            if X_train[0] == "MODE_TEST" or X_train[0] == "MODE_TRAIN_TEST":
                break

            X_train = X_train.to(device)
            X_train.requires_grad = True
            y_train = buffer.take_forward()
            logger.info("Forward received")
            y_train = [y_train[0].to(device), y_train[1].to(device)]
            utility_out = utility_model(X_train)

            loss = loss_fn(utility_out, y_train[0].clone().detach())

            utility_running_loss += loss

            utility_optimizer.zero_grad()
            loss.backward()
            utility_optimizer.step()

            logger.info("Sending gradients")
            buffer.send_backward(X_train.grad.data)

            # ============ Step 2 ============
            logger.info("STEP 2")
            logger.info("Taking forward")
            X_train = buffer.take_forward()
            y_train = buffer.take_forward()
            X_train = X_train.to(device)
            y_train = [y_train[0].to(device), y_train[1].to(device)]
            logger.info("Forward received")

            if len(X_train) == 3:
                privacy_out = privacy_model(X_train[0], X_train[1], X_train[2])
            else:
                privacy_out = privacy_model(X_train)
            loss = loss_fn(privacy_out, y_train[1].clone().detach())

            privacy_running_loss += loss

            privacy_optimizer.zero_grad()
            loss.backward()
            privacy_optimizer.step()

            # ============ Step 3 ============
            logger.info("STEP 3")
            logger.info("Taking forward")
            X_train = buffer.take_forward()
            y_train = buffer.take_forward()
            X_train = X_train.to(device)
            y_train = [y_train[0].to(device), y_train[1].to(device)]
            logger.info("Forward received")

            X_train.requires_grad = True

            # Compute loss and consequent gradients for both networks.
            # Negatively weigh privacy network loss and then update only utility network.
            utility_out = utility_model(X_train)
            privacy_out = privacy_model(X_train)

            utility_loss = loss_fn(utility_out, y_train[0].clone().detach())
            privacy_loss = loss_fn(privacy_out, y_train[1].clone().detach())

            complete_loss = utility_loss + privacy_loss * -1

            utility_optimizer.zero_grad()
            complete_loss.backward()
            utility_optimizer.step()

            logger.info("Sending gradients")
            buffer.send_backward(X_train)

        # Testing after every epoch
        utility_model.eval()
        privacy_model.eval()
        evaluation = {}
        evaluation["Utility/LearningRate"] = utility_scheduler.get_last_lr()
        evaluation["Privacy/LearningRate"] = privacy_scheduler.get_last_lr()
        utility_scheduler.step()
        privacy_scheduler.step()

        logger.warning("Testing")
        evaluation["Utility/Loss/train"] = utility_running_loss / count_running_loss
        evaluation["Privacy/Loss/train"] = privacy_running_loss / count_running_loss

        utility_correct_pred = 0
        privacy_correct_pred = 0
        all_pred = 0
        user_true = []
        user_pred = []
        wandb_report = False
        while True:
            X_test = buffer.take_forward()
            if X_test[0][0] == "MODE_TEST" or X_test[0] == "MODE_TEST":
                break
            # wandb_report = True
            y_test = buffer.take_forward()
            X_test = X_test.to(device)
            y_test = [y_test[0].to(device), y_test[1].to(device)]
            if len(X_test) == 3:
                utility_out = utility_model(X_test[0], X_test[1], X_test[2])
            else:
                utility_out = utility_model(X_test)

            real_users = argmax(y_test[0].clone().detach(), dim=1)
            pred_users = argmax(utility_out, dim=1)
            utility_correct_pred += torch.sum((real_users == pred_users))

            user_true.extend(real_users.cpu())
            user_pred.extend(pred_users.cpu())

            if len(X_test) == 3:
                privacy_out = privacy_model(X_test[0], X_test[1], X_test[2])
            else:
                privacy_out = privacy_model(X_test)

            privacy_correct_pred += torch.sum(
                (
                    argmax(privacy_out, dim=1)
                    == argmax(y_test[1].clone().detach(), dim=1)
                )
            )

            # We use this when we utilise the custom sampler
            # logger.info("Sending predictions.")
            # buffer.send_backward(user_pred)
            # logger.info("Predictions sent.")

            all_pred += len(utility_out)
        if wandb_report:
            logger.warning(
                "Epoch utility train accuracy: {}".format(
                    utility_correct_pred / all_pred
                )
            )
            evaluation["Utility/Accuracy/train"] = utility_correct_pred / all_pred * 100
            logger.warning(
                "Epoch privacy train accuracy: {}".format(
                    privacy_correct_pred / all_pred
                )
            )
            evaluation["Privacy/Accuracy/train"] = privacy_correct_pred / all_pred * 100
            # writer.add_figure(
            #     "Utility/User/Accuracy/train",
            #     generate_confusion_matrix(user_true, user_pred),
            #     epoch,
            # )

        utility_correct_pred = 0
        privacy_correct_pred = 0
        all_pred = 0
        user_true = []
        user_pred = []
        while True:
            X_test = buffer.take_forward()
            if X_test[0][0] == "MODE_TRAIN" or X_test[0] == "MODE_TRAIN":
                break
            wandb_report = True
            y_test = buffer.take_forward()
            X_test = X_test.to(device)
            y_test = [y_test[0].to(device), y_test[1].to(device)]

            if len(X_test) == 3:
                utility_out = utility_model(X_test[0], X_test[1], X_test[2])
            else:
                utility_out = utility_model(X_test)
            real_users = argmax(y_test[0].clone().detach(), dim=1)
            pred_users = argmax(utility_out, dim=1)
            utility_correct_pred += torch.sum((real_users == pred_users))

            user_true.extend(real_users.cpu())
            user_pred.extend(pred_users.cpu())

            if len(X_test) == 3:
                privacy_out = privacy_model(X_test[0], X_test[1], X_test[2])
            else:
                privacy_out = privacy_model(X_test)
            privacy_correct_pred += torch.sum(
                (
                    argmax(privacy_out, dim=1)
                    == argmax(y_test[1].clone().detach(), dim=1)
                )
            )

            all_pred += len(utility_out)
        if wandb_report:
            logger.warning(
                "Epoch utility test accuracy: {}".format(
                    utility_correct_pred / all_pred
                )
            )
            evaluation["Utility/Accuracy/test"] = utility_correct_pred / all_pred * 100
            logger.warning(
                "Epoch privacy test accuracy: {}".format(
                    privacy_correct_pred / all_pred
                )
            )
            evaluation["Privacy/Accuracy/test"] = privacy_correct_pred / all_pred * 100

        running_times.append(datetime.now() - start_time)
        logger.warning("Epoch running time: {}".format(datetime.now() - start_time))
        logger.warning(
            "Mean running time (excluding first epoch): {}".format(
                mean(running_times[1:])
            )
        )

        # Report metrics to wandb
        if wandb_report:
            wandb.log(evaluation)

            # # Save model
            # model_handler.save_model(
            #     "utility_model.pt",
            #     utility_model,
            #     utility_optimizer,
            #     epoch,
            #     utility_correct_pred / all_pred,
            #     privacy_correct_pred / all_pred,
            # )
            # model_handler.save_model(
            #     "privacy_model.pt",
            #     privacy_model,
            #     privacy_optimizer,
            #     epoch,
            #     utility_correct_pred / all_pred,
            #     privacy_correct_pred / all_pred,
            # )

    wandb.finish()
    logger.warning("Server training completed.")
    print(mean(running_times[1:]))
